@@ -16,23 +16,19 @@
  *
  * You should have received a copy of the GNU Lesser General Public
  * License along with the GNU C Library; if not, write to the Free
- * Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
- * 02111-1307 USA.
+ * Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
+ * 02110-1301 USA.
  */
 
-#ifndef G_CANONICALIZE_H
-#define G_CANONICALIZE_H
+#ifndef MATE_CANONICALIZE_H
+#define MATE_CANONICALIZE_H
 
 #include <glib.h>
 
-#ifdef __cplusplus
-extern "C" {
-#endif
+G_BEGIN_DECLS
 
 char* menu_canonicalize_file_name(const char* name, gboolean allow_missing_basename);
 
-#ifdef __cplusplus
-}
-#endif
+G_END_DECLS
 
-#endif /* G_CANONICALIZE_H */
+#endif /* MATE_CANONICALIZE_H */
